#include "gens.h"
#include <readline/readline.h>
#include <stdlib.h>
#include <string.h>
#include "mem_m68k.h"
#include "mem_s68k.h"
#include "mem_sh2.h"
#include "mem_z80.h"
#include "ram_search.h"

//really naive and inefficient, but also simple
int good[RAM_SIZE];
int prev[RAM_SIZE];
int curr[RAM_SIZE];
int show_ram[RAM_VALS_SHOW];
char* show_ram_comments[RAM_VALS_SHOW];
int i_good;
int started = 0;

void RAM_Search_Menu()
{
    if (!started)
        for(int i=0; i<RAM_VALS_SHOW;i++)
            show_ram_comments[i] = NULL;
    started = 1;

    char *tmp_str;
    int keep_searching = 1;
    printf("RAM search menu:\nh for help\n");
    while (keep_searching)
    {
        tmp_str = readline(">");
        if (tmp_str[0] == 'h')
        {
            printf("h - help\n");
            printf("n - new search\n");
            printf("d - display current results\n");
            printf("p - prune current results\n");
            printf("    [+-<>=!@][0-9]+\n");
            printf("u - update current results\n");
            printf("    [cp]\n");
            printf("v - display result on screen\n");
            printf("    [0-9]\n");
            printf("a - display all results on screen\n");
            printf("c - clear all results from screen\n");
            printf("r - read values from file\n");
            printf("    file.grv\n");
            printf("q - quit search\n");
        }
        else if (tmp_str[0] == 'n')
        {
            clear_results();
            printf("results cleared\n");
        }
        else if (tmp_str[0] == 'd')
        {
            display_results();
        }
        else if (tmp_str[0] == 'p')
        {
            prompt_prune(tmp_str+1);
        }
        else if (tmp_str[0] == 'u')
        {
            //copy ram into prev
            if ( (strlen(tmp_str) == 2) && (tmp_str[1] == 'p'))
            {
                get_curr();
                printf("updated current results, didn't update previous\n");
            //copy curr to prev, copy ram to curr
            } else {
                update_results();
                printf("current and previous results updated\n");
            }
        }
        else if (tmp_str[0] == 'v')
        {
            if (tmp_str[1] >= '0' && tmp_str[1] <= '8')
            {
                show_ram[tmp_str[1]-'0'] = i_good;
                printf("displaying address %.4x in position %d\n",i_good,tmp_str[1]-'0');
            }
        }
        else if (tmp_str[0] == 'a')
        {
            int i_sr = 0;
            int good_count = 0;
            for(int i = 0; i<RAM_SIZE; i++)
            {
                if (good[i] && good_count <= 10)
                {
                    show_ram[i_sr] = i;
                    printf("displaying address %.4x in position %d\n",i,i_sr);
                    i_sr++;
                    good_count++;
                }
            }
            //clear any remaining slots
            while(i_sr < RAM_VALS_SHOW)
                show_ram[i_sr++] = 0;
        }
        else if (tmp_str[0] == 'c')
        {
            int i = 0;
            while(i < RAM_VALS_SHOW)
                show_ram[i++] = 0;
            printf("all values cleared from screen\n");
        }
        else if (tmp_str[0] == 'r')
        {
            if (strlen(&tmp_str[1]) < 1)
            {
                printf("please specify a file\n");
            }
            else if (!read_ramview_file(&tmp_str[1]))
            {
                printf("couldn't open file\n");
            }
        }
        else if (tmp_str[0] == 'q')
        {
            keep_searching = 0;
        }
    }
}

void prompt_prune(const char* prune_str)
{
    int num = strtol(prune_str+1,NULL,16);
    int elim = 0;
    if (*prune_str == '>')
    {
        if (strlen(prune_str) == 1)
        {
            printf("looking for decreasing values\n");
            elim = find_decreasing();
        }
        else
        {
            printf("looking for values greater than %d\n",num);
            elim = find_greater_than(num);
        }
    }
    else if (*prune_str == '<')
    {
        if (strlen(prune_str) == 1)
        {
            printf("looking for increasing values\n");
            elim = find_increasing();
        }
        else
        {
            printf("looking for values less than %d\n",num);
            elim = find_less_than(num);
        }
    }
    else if (*prune_str == '+')
    {
        if (strlen(prune_str) == 1)
        {
            printf("looking for increasing values\n");
            elim = find_increasing();
        }
        else
        {
            printf("looking for values that increased by %d\n",num);
            elim = find_increased_by(num);
        }
    }
    else if (*prune_str == '-')
    {
        if (strlen(prune_str) == 1)
        {
            printf("looking for decreasing values\n");
            elim = find_decreasing();
        }
        else
        {
            printf("looking for values that decreased by %d\n",num);
            elim = find_decreased_by(num);
        }
    }
    else if (*prune_str == '=')
    {
        if (strlen(prune_str) == 1)
        {
            printf("looking for unchanged values\n");
            elim = find_unchanged();
        }
        else
        {
            printf("looking for values equal to %d\n",num);
            elim = find_equal_to(num);
        }
    }
    else if (*prune_str == '!')
    {
        if (strlen(prune_str) == 1)
        {
            printf("looking for changed values\n");
            elim = find_changed();
        }
        else
        {
            printf("looking for values not equal to %d\n",num);
            elim = find_not_equal_to(num);
        }
    }
    else if (*prune_str == '@')
    {
        if (strlen(prune_str) == 1)
        {
            printf("you need to specify a memory location\n");
        }
        else
        {
            printf("looking for value at location %d\n",num);
            elim = find_value_at(num);
        }
    }
    else
    {
        printf("command not recognized\n");
    }
    printf("elimintated %d values\n",elim);
    display_results();
}

void update_results()
{
    int i = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        prev[i] = curr[i];
    }
    get_curr();
}

void display_results()
{
    //find number of uneliminated results
    int result_count = 0;
    int i = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        result_count += good[i];
        if (good[i])
            i_good = i;
    }
    if (result_count == 1)
    {
        get_curr();
        printf("0x%.4x is your address!\n",i_good);
        printf("%d is its value!\n",curr[i_good]);
        /*for(int i = 0; i<RAM_VALS_SHOW; i++)
        {
            if (show_ram[i] == 0)
            {
                show_ram[i] = i_good;
                printf("displaying in position %d\n",i);
                i = RAM_VALS_SHOW;
            }
        }*/
    }
    else if (result_count <= RAM_VALS_SHOW)
    {
        printf("%d results:\n",result_count);
        for(int i=0; i<RAM_SIZE; i++)
        {
            if (good[i])
                printf("%.4x matches\n",i);
        }
    }
    else
    {
        printf("%d results\n",result_count);
    }
}

int read_ramview_file(char* filename)
{
    //try to open file
    if (strlen(filename) == 0)
        return 0;
    char* buf = NULL;
    char comment[1024];
    for(int i=0; i<1024;i++)
        comment[i] = 0;
    size_t size = 1024;
    int addr = 0;
    int pos = 0;
    FILE* f = fopen(filename,"r");

    if (f == NULL)
        return 0;
    //read a line
    while (getline(&buf,&size,f) > -1)
    {
        addr = -1;
        pos = -1;
        sscanf(buf,"%d:%x %1023c",&pos,&addr,(char*)&comment);
        //chop off newline
        comment[strlen(comment)-1] = 0;
        if (addr > -1 && pos > -1)
        {
            /*if (show_ram_comments[pos])
                free(show_ram_comments[pos]);
            show_ram_comments[pos] = (char*)malloc(strlen(comment)+1);
            strncpy(show_ram_comments[pos],comment,1023);*/
            show_ram[pos] = addr;
            printf("showing address %.4x in position %d (%s)\n",addr,pos,comment);
        }
        for(int i=0; i<1024;i++)
            comment[i] = 0;
    }

    return 1;
}

void free_ramview_comments()
{
    /*for(int i=0; i<RAM_VALS_SHOW; i++)
    {
        printf("freeing #%d?\n",i);
        if (show_ram_comments[i])
        {
            printf("yes!\n");
            free(show_ram_comments[i]);
        }
        show_ram_comments[i] = NULL;
    }*/
}

void clear_results()
{
    int i = 0;
    for (i=0; i < RAM_SIZE; i++)
    {
        good[i] = 1;
        prev[i] = 0;
    }
    get_curr();
}

void get_curr()
{
    int i = 0;
    for (i=0; i < RAM_SIZE; i++)
    {
        curr[i] = RAM_VAL(i);
    }
}

//I love copy/paste
int find_increasing()
{
    int i = 0,elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((prev[i] < curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            elim++;
            good[i] = 0;
        }
    }
    return elim;
}

int find_less_than(int x)
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((x > curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}

int find_greater_than(int x)
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((x < curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}

int find_decreasing()
{
    int i = 0,elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((prev[i] > curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}

int find_increased_by(int x)
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((prev[i]+x == curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}

int find_decreased_by(int x)
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((prev[i]-x == curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}

int find_unchanged()
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((prev[i] == curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}


int find_equal_to(int x)
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((x == curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}


int find_changed()
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((prev[i] != curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}


int find_not_equal_to(int x)
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if ((x != curr[i]) && good[i])
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}

int find_value_at(int x)
{
    int i = 0, elim = 0;
    for (i=0; i<RAM_SIZE; i++)
    {
        if (x == i)
        {
            good[i] = 1;
        }
        else if (good[i])
        {
            good[i] = 0;
            elim++;
        }
    }
    return elim;
}

